from .syntax import (Forward, Suppress, FollowedBy, Group, OneOrMore, Optional,
                     opAssoc)
 
 
def myOperatorPrecedence(baseExpr, opList):
    """Helper method for constructing grammars of expressions made up of
       operators working in a precedence hierarchy.  Operators may be unary or
       binary, left- or right-associative.  Parse actions can also be attached
       to operator expressions.
 
       Parameters:
        - baseExpr - expression representing the most basic element for the
        - opList - list of tuples, one for each operator precedence level in 
          expression grammar; each tuple is of the form
          (opExpr, numTerms, rightLeftAssoc, parseAction), where:
           - opExpr is the pyparsing expression for the operator;
              may also be a string, which will be converted to a Literal;
              if numTerms is 3, opExpr is a tuple of two expressions, for the
              two operators separating the 3 terms
           - numTerms is the number of terms for this operator (must
              be 1, 2, or 3)
           - rightLeftAssoc is the indicator whether the operator is
              right or left associative, using the pyparsing-defined
              constants opAssoc.RIGHT and opAssoc.LEFT.
           - parseAction is the parse action to be associated with
              expressions matching this operator expression (the
              parse action tuple member may be omitted)
    """
    ret = Forward()
    allops = [x[0] for x in opList]
    opnames = ",".join(str(x) for x in allops)
    ret.setName('operatorSystem(%s)' % opnames)
    # parenthesis = Suppress('(') + ret + FollowedBy(NotAny(oneOf(allops)))
    # - Suppress(')')
    parenthesis = Suppress('(') - ret - Suppress(')')
    lastExpr = parenthesis.setName('parenthesis(%s)' % opnames) | baseExpr
    #lastExpr.setName('Base operand (%s) or parenthesis' % str(baseExpr))
    for operDef in opList:
        opExpr, arity, rightLeftAssoc, pa = (operDef + (None,))[:4]
        if arity == 3:
            if opExpr is None or len(opExpr) != 2:
                raise ValueError(
                    "if numterms=3, opExpr must be a tuple or list of two "
                    "expressions")
            opExpr1, opExpr2 = opExpr
        thisExpr = Forward().setName("operation_with(%s)" % opExpr)
        if rightLeftAssoc == opAssoc.LEFT:
            if arity == 1:
                matchExpr = FollowedBy(lastExpr + opExpr) + Group(
                    lastExpr + OneOrMore(opExpr))
            elif arity == 2:
                if opExpr is not None:
                    #                    matchExpr = Group(lastExpr +
                    # FollowedBy(opExpr) + OneOrMore(opExpr - lastExpr))

                    matchExpr = Group(
                        lastExpr + FollowedBy(opExpr) - OneOrMore(
                            opExpr - lastExpr))
                else:
                    matchExpr = FollowedBy(lastExpr + lastExpr) + Group(
                        lastExpr + OneOrMore(lastExpr))
            elif arity == 3:
                matchExpr = FollowedBy(
                    lastExpr + opExpr1 + lastExpr + opExpr2 + lastExpr) + \
                            Group(
                                lastExpr + opExpr1 + lastExpr + opExpr2 +
                                lastExpr)
            else:
                raise ValueError(
                    "operator must be unary (1), binary (2), or ternary (3)")
        elif rightLeftAssoc == opAssoc.RIGHT:
            if arity == 1:
                # try to avoid LR with this extra test
                if not isinstance(opExpr, Optional):
                    opExpr = Optional(opExpr)
                matchExpr = FollowedBy(opExpr.expr - thisExpr) - Group(
                    opExpr - thisExpr)
            elif arity == 2:
                if opExpr is not None:
                    matchExpr = FollowedBy(
                        lastExpr + opExpr - thisExpr) + Group(
                        lastExpr + OneOrMore(opExpr - thisExpr))
                else:
                    matchExpr = FollowedBy(lastExpr + thisExpr) + Group(
                        lastExpr + OneOrMore(thisExpr))
            elif arity == 3:
                matchExpr = FollowedBy(
                    lastExpr + opExpr1 + thisExpr + opExpr2 + thisExpr) + \
                            Group(
                                lastExpr + opExpr1 + thisExpr + opExpr2 +
                                thisExpr)
            else:
                raise ValueError(
                    "operator must be unary (1), binary (2), or ternary (3)")
        else:
            raise ValueError(
                "operator must indicate right or left associativity")
        if pa:
            matchExpr.setParseAction(pa)
        thisExpr << (matchExpr | lastExpr).setName(
            'operation with %s' % opExpr)
        lastExpr = thisExpr
    ret << lastExpr
    return ret



from collections import deque
from functools import partial
from itertools import (chain, combinations, cycle, dropwhile, islice, repeat,
                       starmap, takewhile, tee)
from operator import add, attrgetter, itemgetter
from sys import version_info

from .func import F
from .op import flip
from .uniform import filterfalse, zip_longest, Iterable


def take(limit, base):
    return islice(base, limit)


def drop(limit, base):
    return islice(base, limit, None)


def takelast(n, iterable):
    "Return iterator to produce last n items from origin"
    return iter(deque(iterable, maxlen=n))


def droplast(n, iterable):
    "Return iterator to produce items from origin except last n"
    t1, t2 = tee(iterable)
    return map(itemgetter(0), zip(t1, islice(t2, n, None)))


def consume(iterator, n=None):
    """Advance the iterator n-steps ahead. If n is none, consume entirely.

    http://docs.python.org/3.4/library/itertools.html#itertools-recipes
    """
    # Use functions that consume iterators at C speed.
    if n is None:
        # feed the entire iterator into a zero-length deque
        deque(iterator, maxlen=0)
    else:
        # advance to the empty slice starting at position n
        next(islice(iterator, n, n), None)


def nth(iterable, n, default=None):
    """Returns the nth item or a default value

    http://docs.python.org/3.4/library/itertools.html#itertools-recipes
    """
    return next(islice(iterable, n, None), default)


def first_true(iterable, default=False, pred=None):
    """Returns the first true value in the iterable.
    If no true value is found, returns *default*
    http://docs.python.org/3.4/library/itertools.html#itertools-recipes
    """
    return next(filter(pred, iterable), default)


# widely-spreaded shortcuts to get first item, all but first item,
# second item, and first item of first item from iterator respectively
head = first = partial(flip(nth), 0)
tail = rest = partial(drop, 1)
second = F(rest) >> first
ffirst = F(first) >> first

# shortcut to remove all falsey items from iterable
compact = partial(filter, None)

# filterfalse under alias 'reject'
reject = filterfalse

# shortcuts to 1. return True if f(x) is logical true for every x in
# iterable (False otherwise), and 2. return the first logical true
# value of f(x) for any x in iterable (None otherwise) respectively
every = F(partial(map)) >> all
some = F(partial(map)) >> compact >> first


def iterate(f, x):
    """Return an iterator yielding x, f(x), f(f(x)) etc.
    """
    while True:
        yield x
        x = f(x)


def padnone(iterable):
    """Returns the sequence elements and then returns None indefinitely.
    Useful for emulating the behavior of the built-in map() function.

    http://docs.python.org/3.4/library/itertools.html#itertools-recipes
    """
    return chain(iterable, repeat(None))


def ncycles(iterable, n):
    """Returns the sequence elements n times

    http://docs.python.org/3.4/library/itertools.html#itertools-recipes
    """
    return chain.from_iterable(repeat(tuple(iterable), n))


def repeatfunc(func, times=None, *args):
    """Repeat calls to func with specified arguments.
    Example:  repeatfunc(random.random)

    http://docs.python.org/3.4/library/itertools.html#itertools-recipes
    """
    if times is None:
        return starmap(func, repeat(args))
    return starmap(func, repeat(args, times))


def grouper(n, iterable, fillvalue=None):
    """Collect data into fixed-length chunks or blocks, so
    grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx

    http://docs.python.org/3.4/library/itertools.html#itertools-recipes
    """
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)

def group_by(keyfunc, iterable):
    """Returns a dict of the elements from given iterable keyed by result
    of keyfunc on each element. The value at each key will be a list of
    the corresponding elements, in the order they appeared in the iterable.
    """
    grouped = {}
    for item in iterable:
        grouped.setdefault(keyfunc(item), []).append(item)
    return grouped


def roundrobin(*iterables):
    """roundrobin('ABC', 'D', 'EF') --> A D E B F C
    Recipe originally credited to George Sakkis.
    Reimplemented to work both in Python 2+ and 3+.

    http://docs.python.org/3.4/library/itertools.html#itertools-recipes
    """
    pending = len(iterables)
    next_attr = "next" if version_info[0] == 2 else "__next__"
    nexts = cycle(map(attrgetter(next_attr), map(iter, iterables)))
    while pending:
        try:
            for n in nexts:
                yield n()
        except StopIteration:
            pending -= 1
            nexts = cycle(islice(nexts, pending))


def partition(pred, iterable):
    """Use a predicate to partition entries into false entries and true entries
    partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9

    http://docs.python.org/3.4/library/itertools.html#itertools-recipes
    """
    t1, t2 = tee(iterable)
    return filterfalse(pred, t1), filter(pred, t2)


def splitat(t, iterable):
    """Split iterable into two iterators after given number of iterations
    splitat(2, range(5)) --> 0 1 and 2 3 4
    """
    t1, t2 = tee(iterable)
    return islice(t1, t), islice(t2, t, None)


def splitby(pred, iterable):
    """Split iterable into two iterators at first false predicate
    splitby(is_even, range(5)) --> 0 and 1 2 3 4
    """
    t1, t2 = tee(iterable)
    return takewhile(pred, t1), dropwhile(pred, t2)


def powerset(iterable):
    """powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)

    http://docs.python.org/3.4/library/itertools.html#itertools-recipes
    """
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))


def pairwise(iterable):
    """pairwise(s) -> (s0,s1), (s1,s2), (s2, s3), ...

    http://docs.python.org/3.4/library/itertools.html#itertools-recipes
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def iter_suppress(func, exception, first_=None):
    """ Call a function repeatedly until an exception is raised.

    Converts a call-until-exception interface to an iterator interface.
    Like __builtin__.iter(func, sentinel) but uses an exception instead
    of a sentinel to end the loop.

    Examples:
        # priority queue iterator
        iter_except(functools.partial(heappop, h), IndexError)

        # non-blocking dict iterator
        iter_except(d.popitem, KeyError)

        # non-blocking deque iterator
        iter_except(d.popleft, IndexError)

        # loop over a producer Queue
        iter_except(q.get_nowait, Queue.Undefined)

        # non-blocking set iterator
        iter_except(s.pop, KeyError)

    http://docs.python.org/3.4/library/itertools.html#itertools-recipes
    """
    try:
        # For database APIs needing an initial cast to db.first()
        if first_ is not None:
            yield first_()
        while 1:
            yield func()
    except exception:
        pass


def flatten(items):
    """Flatten any level of nested iterables (not including strings, bytes or
    bytearrays).
    Reimplemented to work with all nested levels (not only one).

    http://docs.python.org/3.4/library/itertools.html#itertools-recipes
    """
    str_type = basestring if version_info[0] < 3 else str
    for item in items:
        is_iterable = isinstance(item, Iterable)
        is_string_or_bytes = isinstance(item, (str_type, bytes, bytearray))
        if is_iterable and not is_string_or_bytes:
            for i in flatten(item):
                yield i
        else:
            yield item


if version_info[0] == 3 and version_info[1] >= 3:
    from itertools import accumulate



from functools import partial, update_wrapper, wraps
from sys import version_info
from symtable import Function

_has_type_hint_support = version_info[:2] >= (3, 5)


def identity(arg):
    return arg


class F(Function):
    """Provide simple syntax for functions composition
    (through << and >> operators) and partial function
    application (through simple tuple syntax).

    Usage example:

    >>> func = F() << (_ + 10) << (_ + 5)
    >>> print(func(10))
    25
    >>> func = F() >> (filter, _ < 6) >> sum
    >>> print(func(range(10)))
    15
    """

    __slots__ = "f",

    def __init__(self, f=identity, *args, **kwargs):
        super(F, self).__init__(self, 'F')
        self.f = partial(f, *args, **kwargs) if any([args, kwargs]) else f

    @classmethod
    def __compose(cls, f, g):
        """Produces new class instance that will
        execute given functions one by one. Internal
        method that was added to avoid code duplication
        in other methods.
        """
        return cls(lambda *args, **kwargs: f(g(*args, **kwargs)))

    def __ensure_callable(self, f):
        """Simplify partial execution syntax.
        Rerurn partial function built from tuple
        (func, arg1, arg2, ...)
        """
        return self.__class__(*f) if isinstance(f, tuple) else f

    def __rshift__(self, g):
        """Overload >> operator for F instances"""
        return self.__class__.__compose(self.__ensure_callable(g), self.f)

    def __lshift__(self, g):
        """Overload << operator for F instances"""
        return self.__class__.__compose(self.f, self.__ensure_callable(g))

    def __call__(self, *args, **kwargs):
        """Overload apply operator"""
        return self.f(*args, **kwargs)


def curried(func):
    """A decorator that makes the function curried

    Usage example:

    >>> @curried
    ... def sum5(a, b, c, d, e):
    ...     return a + b + c + d + e
    ...
    >>> sum5(1)(2)(3)(4)(5)
    15
    >>> sum5(1, 2, 3)(4, 5)
    15
    """

    def _args_len(func):
        if _has_type_hint_support:
            from inspect import signature
            args = signature(func).parameters
        else:  # pragma: no cover
            from inspect import getargspec
            args = getargspec(func).args

        return len(args)

    @wraps(func)
    def _curried(*args, **kwargs):
        f = func
        count = 0
        while isinstance(f, partial):
            if f.args:
                count += len(f.args)
            f = f.func

        if count == _args_len(f) - len(args):
            return func(*args, **kwargs)

        para_func = partial(func, *args, **kwargs)
        update_wrapper(para_func, f)
        return curried(para_func)

    return _curried



from hypothesis import given
from hypothesis import strategies as st
from Aspidites._vendor.fn import Stream
from Aspidites._vendor.fn import iters


@given(st.lists(st.integers(), min_size=1))
def test_from_list(list_: list):
    s = Stream() << list_
    assert list_ == list(s)
    s = Stream() << iter(list_)
    assert list_ == list(s)


def test_from_generator():
    def gen():
        yield 1
        yield 2
        yield 3

    s = Stream() << gen << (4, 5)
    assert list(s) == [1, 2, 3, 4, 5]


def test_lazy_slicing():
    s = Stream() << range(10)
    assert s.cursor() == 0

    s_slice = s[:5]
    assert s.cursor() == 0
    assert len(list(s_slice)) == 5


def test_lazy_slicing_recursive():
    s = Stream() << range(10)
    sf = s[1:3][0:2]

    assert s.cursor() == 0
    assert len(list(sf)) == 2


def test_fib_infinite_stream():
    from operator import add

    f = Stream()
    fib = f << [0, 1] << map(add, f, iters.drop(1, f))

    assert [0, 1, 1, 2, 3, 5, 8, 13, 21, 34] == list(iters.take(10, fib))
    assert 6765 == fib[20]
    assert [832040, 1346269, 2178309, 3524578, 5702887] == list(fib[30:35])
    # 35 elements should be already evaluated
    assert fib.cursor() == 35


def test_origin_param():
    assert [100] == list(Stream(100))
    assert [1, 2, 3] == list(Stream(1, 2, 3))
    assert [1, 2, 3, 10, 20, 30] == list(Stream(1, 2, 3) << [10, 20, 30])


@given(st.text())
def test_origin_param_string(s):
    assert [s] == list(Stream(s))



from . import syntax_fail, good, fail, semantic_fail

# basic types

good('int', 1)
fail('int', None)
fail('int', 2.0)
good('float', 1.1)
fail('float', None)
fail('float', 2)

good('number', 1)
good('number', 1.0)
fail('number', [1])

good('bool', False)
good('bool', True)
fail('bool', 1)
fail('bool', 0)

# type contract

good('None', None)
good('NoneType', None)
fail('None', 1)
fail('NoneType', 1)

syntax_fail('type')
syntax_fail('type()')
good('type(x)', 1)
semantic_fail('type(X)', 1)



import operator
from hypothesis import strategies as st
from hypothesis import given
from Aspidites._vendor.fn import F, _, curried


def test_curried():
    @curried
    def sum5(a, b, c, d, e):
        return a + b + c + d + e
    assert sum5(1)(2)(3)(4)(5) == sum5(1, 2, 3)(4, 5)


@given(st.integers())
def test_composition(x):
    def f(x_):
        return x_ * 2

    def g(x_):
        return x_ + 10

    assert (x + 10) * 2 == (F(f) << g)(x)

    def z(x_):
        return x_ * 20
    assert ((x * 20) + 10) * 2 == (F(f) << F(g) << F(z))(x)


@given(st.integers(), st.integers(), st.integers())
def test_partial(x, y, z):
    # Partial should work if we pass additional arguments to F constructor
    f = F(operator.add, y) << F(operator.add, z)
    assert x + y + z == f(x)


def test_underscore():
    assert [1, 4, 9] == list(map(F() << (_ ** 2) << _ + 1, range(3)))


@given(st.integers())
def test_pipe_composition(x):
    def f(x_):
        return x_ * 2

    def g(x_):
        return x_ + 10

    assert (x * 2) + 10 == (F() >> f >> g)(x)


def test_pipe_partial():
    func = F() >> (filter, _ < 6) >> sum
    assert 15 == func(range(10))



import pytest as pt

from Aspidites._vendor.fn import _
from Aspidites._vendor.fn import underscore
from Aspidites._vendor.fn import reduce



def test_identity_default():
    assert 10, _(10)

def test_arithmetic():
    # operator +
    assert 7 == (_ + 2)(5)
    assert [10, 11, 12], list(map(_ + 10, [0, 1, 2]))
    # operator -
    assert 3 == (_ - 2)(5)
    assert 13 == (_ - 2 + 10)(5)
    assert [0, 1, 2], list(map(_ - 10, [10, 11, 12]))
    # operator *
    assert 10 == (_ * 2)(5)
    assert 50 == (_ * 2 + 40)(5)
    assert [0, 10, 20], list(map(_ * 10, [0, 1, 2]))
    # operator /
    assert 5 == (_ / 2)(10)
    assert 6 == (_ / 2 + 1)(10)
    assert [1, 2, 3], list(map(_ / 10, [10, 20, 30]))
    # operator **
    assert 100 == (_ ** 2)(10)
    # operator %
    assert 1 == (_ % 2)(11)
    # operator <<
    assert 32 == (_ << 2)(8)
    # operator >>
    assert 2 == (_ >> 2)(8)
    # operator (-a)
    assert 10,  (-_)(-10)
    assert -10 == (-_)(10)
    # operator (+a)
    assert 10,  (+_)(10)
    assert -10 == (+_)(-10)
    # operator (~a)
    assert -11 == (~_)(10)

def test_arithmetic_multiple():
    assert 10 == (_ + _)(5, 5)
    assert 0 == (_ - _)(5, 5)
    assert 25 == (_ * _)(5, 5)
    assert 1 == (_ / _)(5, 5)

def test_arithmetic_swap():
    # operator +
    assert 7 == (2 + _)(5)
    assert [10, 11, 12], list(map(10 + _, [0, 1, 2]))
    # operator -
    assert 3 == (8 - _)(5)
    assert 13 == (8 - _ + 10)(5)
    assert [10, 9, 8], list(map(10 - _, [0, 1, 2]))
    # operator *
    assert 10 == (2 * _)(5)
    assert 50 == (2 * _ + 40)(5)
    assert [0, 10, 20], list(map(10 * _, [0, 1, 2]))
    # operator /
    assert 5 == (10 / _)(2)
    assert 6 == (10 / _ + 1)(2)
    assert [10, 5, 2], list(map(100 / _, [10, 20, 50]))
    # operator **
    assert 100 == (10**_)(2)
    # operator %
    assert 1 == (11 % _)(2)
    # operator <<
    assert 32 == (8 << _)(2)
    # operator >>
    assert 2 == (8 >> _)(2)


def test_bitwise():
    # and
    assert bool((_ & 1)(1)) is True
    assert bool((_ & 1)(0)) is False
    assert bool((_ & 0)(1)) is False
    assert bool((_ & 0)(0)) is False
    # or
    assert bool((_ | 1)(1)) is True
    assert bool((_ | 1)(0)) is True
    assert bool((_ | 0)(1)) is True
    assert bool((_ | 0)(0)) is False
    # xor
    assert bool((_ ^ 1)(0)) is True
    assert bool((_ ^ 0)(1)) is True
    assert bool((_ ^ 1)(1)) is False
    assert bool((_ ^ 0)(0)) is False


def test_bitwise_swap():
    # and
    assert bool((1 & _)(1)) is True
    assert bool((1 & _)(0)) is False
    assert bool((0 & _)(1)) is False
    assert bool((0 & _)(0)) is False
    # or
    assert bool((1 | _)(1)) is True
    assert bool((1 | _)(0)) is True
    assert bool((0 | _)(1)) is True
    assert bool((0 | _)(0)) is False
    # xor
    assert bool((1 ^ _)(0)) is True
    assert bool((0 ^ _)(1)) is True
    assert bool((1 ^ _)(1)) is False
    assert bool((0 ^ _)(0)) is False


def test_getattr():
    class GetattrTest(object):
        def __init__(self):
            self.doc = "TestCase"

    assert "TestCase" == _.doc(GetattrTest())
    assert "TestCaseTestCase" == (_.doc * 2)(GetattrTest())
    assert "TestCaseTestCase" == (_.doc + _.doc)(GetattrTest(), GetattrTest())


def test_call_method():
    assert ["test", "case"] == (_.call("split"))("test case")
    assert "str", _.__name__(str)


def test_call_method_args():
    assert ["test", "case"] == (_.call("split", "-"))("test-case")
    assert ["test-case"] == (_.call("split", "-", 0))("test-case")


def test_call_method_kwargs():
    test_dict = {'num': 23}
    _.call("update", num=42)(test_dict)
    assert {'num': 42} == test_dict


def test_comparator():
    assert (_ < 7)(1) is True
    assert (_ < 7)(10) is False
    assert (_ > 20)(25) is True
    assert (_ > 20)(0) is False
    assert (_ <= 7)(6) is True
    assert (_ <= 7)(7) is True
    assert (_ <= 7)(8) is False
    assert (_ >= 7)(8) is True
    assert (_ >= 7)(7) is True
    assert (_ >= 7)(6) is False
    assert (_ == 10)(10) is True
    assert (_ == 10)(9) is False


# noinspection PyCallingNonCallable
def test_none():
    assert (_ == None)(None) is True  # noqa: E711

    class pushlist(list):
        def __lshift__(self, item):
            self.append(item)
            return self

    assert [None] == (_ << None)(pushlist())

# noinspection PyCallingNonCallable
def test_comparator_multiple():
    assert (_ < _)(1, 2) is True
    assert (_ < _)(2, 1) is False
    assert (_ > _)(25, 20) is True
    assert (_ > _)(20, 25) is False
    assert (_ <= _)(6, 7) is True
    assert (_ <= _)(7, 7) is True
    assert (_ <= _)(8, 7) is False
    assert (_ >= _)(8, 7) is True
    assert (_ >= _)(7, 7) is True
    assert (_ >= _)(6, 7) is False
    assert (_ == _)(10, 10) is True
    assert (_ == _)(9, 10) is False

def test_comparator_filter():
    assert [0, 1, 2] == list(filter(_ < 5, [0, 1, 2, 10, 11, 12]))

def test_slicing():
    assert 0 == (_[0])(list(range(10)))
    assert 9 == (_[-1])(list(range(10)))
    assert [3, 4, 5] == (_[3:])(list(range(6)))
    assert [0, 1, 2] == (_[:3])(list(range(10)))
    assert [1, 2, 3] == (_[1:4])(list(range(10)))
    assert [0, 2, 4] == (_[0:6:2])(list(range(10)))


def test_slicing_multiple():
    assert 0 == (_[_])(range(10), 0)
    assert 8 == (_[_ * (-1)])(range(10), 2)


def test_arity_error():
    with pt.raises(underscore.ArityError):
        _(1, 2)
        (_ + _)(1)
    with pt.raises(TypeError):
        _(1, 2)
        (_ + _)(1)


def test_more_than_2_operations():
    assert 12 == (_ * 2 + 10)(1)
    assert 6,  (_ + _ + _)(1, 2, 3)
    assert 10 == (_ + _ + _ + _)(1, 2, 3, 4)
    assert 7,  (_ + _ * _)(1, 2, 3)


def test_string_converting():
    assert "(x1) => x1", str(_)

    assert "(x1) => (x1 + 2)" == str(_ + 2)
    assert "(x1) => (x1 - 2)" == str(_ - 2)
    assert "(x1) => (x1 * 2)" == str(_ * 2)
    assert "(x1) => (x1 / 2)" == str(_ / 2)
    assert "(x1) => (x1 % 2)" == str(_ % 2)
    assert "(x1) => (x1 ** 2)" == str(_ ** 2)

    assert "(x1) => (x1 & 2)" == str(_ & 2)
    assert "(x1) => (x1 | 2)" == str(_ | 2)
    assert "(x1) => (x1 ^ 2)" == str(_ ^ 2)

    assert "(x1) => (x1 >> 2)" == str(_ >> 2)
    assert "(x1) => (x1 << 2)" == str(_ << 2)

    assert "(x1) => (x1 < 2)" == str(_ < 2)
    assert "(x1) => (x1 > 2)" == str(_ > 2)
    assert "(x1) => (x1 <= 2)" == str(_ <= 2)
    assert "(x1) => (x1 >= 2)" == str(_ >= 2)
    assert "(x1) => (x1 == 2)" == str(_ == 2)
    assert "(x1) => (x1 != 2)" == str(_ != 2)
    assert "(x1) => ((x1 * 2) + 1)" == str((_ * 2 + 1))


def test_rigthside_string_converting():
    assert "(x1) => (2 + x1)" == str(2 + _)
    assert "(x1) => (2 - x1)" == str(2 - _)
    assert "(x1) => (2 * x1)" == str(2 * _)
    assert "(x1) => (2 / x1)" == str(2 / _)
    assert "(x1) => (2 % x1)" == str(2 % _)
    assert "(x1) => (2 ** x1)" == str(2 ** _)

    assert "(x1) => (2 & x1)" == str(2 & _)
    assert "(x1) => (2 | x1)" == str(2 | _)
    assert "(x1) => (2 ^ x1)" == str(2 ^ _)

    assert "(x1) => (2 >> x1)" == str(2 >> _)
    assert "(x1) => (2 << x1)" == str(2 << _)


def test_unary_string_converting():
    assert "(x1) => (+x1)" == str(+_)
    assert "(x1) => (-x1)" == str(-_)
    assert "(x1) => (~x1)" == str(~_)


def test_multiple_string_converting():
    assert "(x1, x2) => (x1 + x2)" == str(_ + _)
    assert "(x1, x2) => (x1 * x2)" == str(_ * _)
    assert "(x1, x2) => (x1 - x2)" == str(_ - _)
    assert "(x1, x2) => (x1 / x2)" == str(_ / _)
    assert "(x1, x2) => (x1 % x2)" == str(_ % _)
    assert "(x1, x2) => (x1 ** x2)" == str(_ ** _)

    assert "(x1, x2) => (x1 & x2)" == str(_ & _)
    assert "(x1, x2) => (x1 | x2)" == str(_ | _)
    assert "(x1, x2) => (x1 ^ x2)" == str(_ ^ _)

    assert "(x1, x2) => (x1 >> x2)" == str(_ >> _)
    assert "(x1, x2) => (x1 << x2)" == str(_ << _)

    assert "(x1, x2) => (x1 > x2)" == str(_ > _)
    assert "(x1, x2) => (x1 < x2)" == str(_ < _)
    assert "(x1, x2) => (x1 >= x2)" == str(_ >= _)
    assert "(x1, x2) => (x1 <= x2)" == str(_ <= _)
    assert "(x1, x2) => (x1 == x2)" == str(_ == _)
    assert "(x1, x2) => (x1 != x2)" == str(_ != _)

    assert "(x1, x2) => (((x1 / x2) - 1) * 100)" == str((_ / _ - 1) * 100)


def test_reverse_string_converting():
    assert "(x1, x2, x3) => ((x1 + x2) + x3)" == str(_ + _ + _)
    assert "(x1, x2, x3) => (x1 + (x2 * x3))" == str(_ + _ * _)

    assert "(x1) => (1 + (2 * x1))", str((1 + 2 * _))


def test_multi_underscore_string_converting():
    assert "(x1) => (x1 + '_')", str(_ + "_")
    assert "(x1, x2) => getattr((x1 + x2), '__and_now__')" == str((_ + _).__and_now__)
    assert "(x1, x2) => x1['__name__'][x2]" == str(_['__name__'][_])


def test_repr():
    assert _ / 2 == eval(repr(_ / 2))
    assert _ + _ == eval(repr(_ + _))
    assert _ + _ * _ == eval(repr(_ + _ * _))


def test_repr_parse_str():
    assert '=> ' + _ == eval(repr('=> ' + _))
    assert reduce(lambda f, n: f.format(n), ('({0} & _)',) * 11).format('_') == repr(reduce(_ & _, (_,) * 12))



import inspect
from typing import Any, Iterable, TypeVar, Union

from pyrsistent import PVector

try:
    from typing import GenericMeta
except ImportError:
    from typing import _GenericAlias as GenericMeta

T = TypeVar("T")


class UnderscoreType:
    def __repr__(self):
        return '_'


class HeadType:
    def __repr__(self):
        return 'HEAD'


class TailType:
    def __repr__(self):
        return 'TAIL'


class PaddedValueType:
    def __repr__(self):
        return 'PaddedValue'


class NoDefaultType:
    pass


NoDefault = NoDefaultType()

PaddedValue = PaddedValueType()


class BoxedArgs:
    def __init__(self, obj):
        self.obj = obj

    def get(self):
        return self.obj


def pairwise(l):
    i = 0
    while i < len(l):
        yield l[i], l[i + 1]
        i += 2


def get_lambda_args_error_msg(action, var, err):
    try:
        code = inspect.getsource(action)
        return "Error passing arguments %s here:\n%s\n%s" % (var, code, err)
    except OSError:
        return "Error passing arguments %s:\n%s" % (var, err)


def is_dataclass(value):
    """
    Dataclass support is only enabled in Python 3.7+, or in 3.6 with the `dataclasses` backport installed
    """
    try:
        from dataclasses import is_dataclass
        return is_dataclass(value)
    except ImportError:
        return False


def get_extra(pattern):
    return getattr(pattern, "__extra__", None) or getattr(pattern, "__origin__", None)


def peek(iter_: Iterable[T]) -> T:
    return next(iter(iter_))


def is_newtype(pattern):
    return inspect.isfunction(pattern) and hasattr(pattern, '__supertype__')


def is_generic(pattern):
    return isinstance(pattern, GenericMeta)


def is_union(pattern):
    return isinstance(pattern, Union.__class__) or getattr(pattern, "__origin__", None) == Union


def is_typing_stuff(pattern):
    return pattern == Any or is_generic(pattern) or is_union(pattern) or is_newtype(pattern)


def get_real_type(subtype):
    if is_newtype(subtype):
        return get_real_type(subtype.__supertype__)
    else:
        return subtype
